@@ -1,11 +1,13 @@
-{-# LANGUAGE CPP                        #-}
 {-# LANGUAGE BangPatterns               #-}
+{-# LANGUAGE CPP                        #-}
+{-# LANGUAGE FlexibleContexts           #-}
+{-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE ImpredicativeTypes         #-}
 {-# LANGUAGE MultiParamTypeClasses      #-}
 {-# LANGUAGE OverloadedStrings          #-}
 {-# LANGUAGE TemplateHaskell            #-}
-{-# LANGUAGE FlexibleInstances          #-}
+{-# LANGUAGE TypeFamilies               #-}
 
 #ifndef MIN_VERSION_comonad
 #define MIN_VERSION_comonad(x,y,z) 1
@@ -13,23 +15,30 @@
 
 module Snap.Snaplet.Internal.Types where
 
-import           Control.Applicative
-import           Control.Lens
-import           Control.Monad.CatchIO hiding (Handler)
-import           Control.Monad.Reader
-import           Control.Monad.State.Class
-import           Control.Monad.Trans.Either
-import           Control.Monad.Trans.Writer hiding (pass)
-import           Data.ByteString (ByteString)
-import qualified Data.ByteString.Char8 as B
-import           Data.Configurator.Types
-import           Data.IORef
-import           Data.Monoid
-import           Data.Text (Text)
+------------------------------------------------------------------------------
+import           Control.Applicative          (Alternative)
+import           Control.Lens                 (ALens', makeLenses, set)
+import           Control.Monad                (liftM)
+import           Control.Monad.Base           (MonadBase (..))
+import           Control.Monad.Reader         (MonadIO (..), MonadPlus, MonadReader (ask, local))
+import           Control.Monad.State.Class    (MonadState (get, put), gets)
+import           Control.Monad.Trans.Control  (MonadBaseControl (..))
+import           Control.Monad.Trans.Writer   (WriterT)
+import           Data.ByteString              (ByteString)
+import qualified Data.ByteString.Char8        as B (dropWhile, intercalate, null, reverse)
+import           Data.Configurator.Types      (Config)
+import           Data.IORef                   (IORef)
+import           Data.Text                    (Text)
+import           Snap.Core                    (MonadSnap, Request (rqClientAddr), Snap, bracketSnap, getRequest, pass, writeText)
+import qualified Snap.Snaplet.Internal.Lensed as L (Lensed (..), runLensed, with, withTop)
+import qualified Snap.Snaplet.Internal.LensT  as LT (LensT, getBase, with, withTop)
 
-import           Snap.Core
-import qualified Snap.Snaplet.Internal.LensT as LT
-import qualified Snap.Snaplet.Internal.Lensed as L
+#if !MIN_VERSION_base(4,8,0)
+import           Control.Applicative          (Applicative)
+import           Data.Monoid                  (Monoid (mappend, mempty))
+#endif
+
+------------------------------------------------------------------------------
 
 
 ------------------------------------------------------------------------------
@@ -121,8 +130,17 @@ snapletConfig :: SimpleLens (Snaplet a) SnapletConfig
 snapletValue :: SimpleLens (Snaplet a) a
 -}
 
+
+-- NOTE: We cannot use one of the smaller lens packages because none of them
+-- include ALens'.  We have to use ALens' because we use lenses inside f's...
+-- f (Lens a b).  That requires ImpredicativeTypes which doesn't work.  We
+-- also can't inline the type aliases because ALens' uses Pretext which is a
+-- newtype and can't be supplied outside lens in a compatible way.
+
+------------------------------------------------------------------------------
 type SnapletLens s a = ALens' s (Snaplet a)
 
+
 ------------------------------------------------------------------------------
 -- | Transforms a lens of the type you get from makeLenses to an similar lens
 -- that is more suitable for internal use.
@@ -251,17 +269,37 @@ snapletURL suffix = do
 -- 'MonadSnaplet' instance, which gives you all the functionality described
 -- above.
 newtype Handler b v a =
-    Handler (L.Lensed (Snaplet b) (Snaplet v) Snap a)
+    Handler { _unHandler :: L.Lensed (Snaplet b) (Snaplet v) Snap a }
   deriving ( Monad
            , Functor
            , Applicative
            , MonadIO
            , MonadPlus
-           , MonadCatchIO
            , Alternative
            , MonadSnap)
 
 
+------------------------------------------------------------------------------
+instance MonadBase IO (Handler b v) where
+    liftBase = liftIO
+
+
+------------------------------------------------------------------------------
+newtype StMHandler b v a = StMHandler {
+      unStMHandler :: StM (L.Lensed (Snaplet b) (Snaplet v) Snap) a
+    }
+
+
+instance MonadBaseControl IO (Handler b v) where
+    type StM (Handler b v) a = StMHandler b v a
+    liftBaseWith f = Handler
+                       $ liftBaseWith
+                       $ \g' -> f
+                       $ \m -> liftM StMHandler
+                       $ g' $ _unHandler m
+    restoreM = Handler . restoreM . unStMHandler
+
+
 ------------------------------------------------------------------------------
 -- | Gets the @Snaplet v@ from the current snaplet's state.
 getSnapletState :: Handler b v (Snaplet v)
@@ -292,14 +330,16 @@ getsSnapletState f = do
 
 
 ------------------------------------------------------------------------------
--- | The MonadState instance gives you access to the current snaplet's state.
+-- | Lets you access the current snaplet's state through the 'MonadState'
+-- interface.
 instance MonadState v (Handler b v) where
     get = getsSnapletState _snapletValue
     put v = modifySnapletState (set snapletValue v)
 
 
 ------------------------------------------------------------------------------
--- | The MonadState instance gives you access to the current snaplet's state.
+-- | Lets you access the current snaplet's state through the 'MonadReader'
+-- interface.
 instance MonadReader v (Handler b v) where
     ask = getsSnapletState _snapletValue
     local f m = do
@@ -310,6 +350,7 @@ instance MonadReader v (Handler b v) where
         return res
 
 
+------------------------------------------------------------------------------
 instance MonadSnaplet Handler where
     getLens = Handler ask
     with' !l (Handler !m) = Handler $ L.with l m
@@ -342,17 +383,22 @@ setRoutePattern p = withTop' id $
     modifySnapletState (set (snapletConfig . scRoutePattern) (Just p))
 
 
+------------------------------------------------------------------------------
+-- | Check whether the request comes from localhost.
+isLocalhost :: MonadSnap m => m Bool
+isLocalhost = do
+    rip <- liftM rqClientAddr getRequest
+    return $ elem rip [ "127.0.0.1"
+                      , "localhost"
+                      , "::1" ]
+
+
 ------------------------------------------------------------------------------
 -- | Pass if the request is not coming from localhost.
 failIfNotLocal :: MonadSnap m => m b -> m b
 failIfNotLocal m = do
-    -- FIXME: this moves to auth once control-panel is done
-    rip <- liftM rqRemoteAddr getRequest
-    if not $ elem rip [ "127.0.0.1"
-                      , "localhost"
-                      , "::1" ]
-      then pass
-      else m
+    isLocal <- isLocalhost
+    if isLocal then m else pass
 
 
 ------------------------------------------------------------------------------
@@ -419,12 +465,21 @@ data InitializerState b = InitializerState
 ------------------------------------------------------------------------------
 -- | Wrapper around IO actions that modify state elements created during
 -- initialization.
-newtype Hook a = Hook (Snaplet a -> EitherT Text IO (Snaplet a))
+newtype Hook a = Hook (Snaplet a -> IO (Either Text (Snaplet a)))
 
 
+------------------------------------------------------------------------------
 instance Monoid (Hook a) where
-    mempty = Hook return
-    (Hook a) `mappend` (Hook b) = Hook (a >=> b)
+    mempty = Hook (return . Right)
+    (Hook a) `mappend` (Hook b) = Hook $ \s -> do
+      ea <- a s
+      case ea of
+        Left e -> return $ Left e
+        Right ares -> do
+          eb <- b ares
+          case eb of
+            Left e -> return $ Left e
+            Right bres -> return $ Right bres
 
 
 ------------------------------------------------------------------------------
@@ -440,6 +495,7 @@ newtype Initializer b v a =
 makeLenses ''InitializerState
 
 
+------------------------------------------------------------------------------
 instance MonadSnaplet Initializer where
     getLens = Initializer ask
     with' !l (Initializer !m) = Initializer $ LT.with l m
@@ -451,5 +507,3 @@ instance MonadSnaplet Initializer where
 -- | Opaque newtype which gives us compile-time guarantees that the user is
 -- using makeSnaplet and either nestSnaplet or embedSnaplet correctly.
 newtype SnapletInit b v = SnapletInit (Initializer b v (Snaplet v))
-
-
